//
//  DTAG100_iOS_example.m
//
//  This file contains extracts from an iOS app (without the distration of the user interface elements)
//  in order to demonstrate the various stages of interacting with the BLE functionality of the DTAG100 
//  using the iOS CoreLocation and CoreBluetooth frameworks.
//
//  This includes using CoreLocation for monitoring and ranging and using CoreBluetooth to discovering 
//  the DTAG100 as a peripheral as well as connecting, discovering GATT services and characteristics and 
//  reading and writing GATT characteristics.
//
//  Copyright (c) 2014 Dot Origin. All rights reserved.
//

@import CoreLocation;
@import CoreBluetooth;

@interface DTAG100ExampleController () <CLLocationManagerDelegate, CBCentralManagerDelegate, CBPeripheralDelegate>

@property (nonatomic) CLLocationManager *locationManager;
@property (nonatomic) CBCentralManager *centralManager;
@property (nonatomic) CBPeripheral *peripheral;

// An array of the beacon UUIDs that we want to monitor
@property NSArray *beaconUUIDs;

// An array of region names corresponding to the beacons UUIDs that we want to monitor
@property NSArray *beaconIdentifiers;

// An array of beacon regions that we are monitoring and possibly ranging
@property NSMutableArray *regions;

// An array of UUIDs (as NSStrings) of the read-only GATT characteristics exposed by the DTAG100
@property NSArray *readCharacteristicUUIDs;

// An array to read and assemble the values of the read-only characteristics read from the DTAG100
@property NSMutableArray *readCharacteristicValues;

// State properties and flags
@property BOOL cBReady;
@property BOOL scanning;
@property uint characteristicsCount;
@property uint readCount;
@property BOOL writeData;

// Properties used to determine how long we have been within range of a beacon
@property BOOL waitingToTrigger;
@property BOOL triggered;
@property NSDate *timeOfEntry;

@end


@implementation DTAG100ExampleController

- (void)viewDidLoad
{
    [super viewDidLoad];
    
	// Set initial values of all properties
    self.cBReady = false;
    self.scanning = false;
    self.readCount = 0;
    self.writeData = false;
	self.waitingToTrigger = false;
	self.triggered = false;
    
	// Initialise a location manager instance and set its delegate to this class instance
    self.locationManager = [[CLLocationManager alloc] init];
    self.locationManager.delegate = self;
    
    self.centralManager = [[CBCentralManager alloc] initWithDelegate:self queue:nil options:nil];
        
	// The DTAG100's custom GATT service UUID 
	// (we will search for peripheral devices that advertise this service)
    self.ourServiceUUID = [CBUUID UUIDWithString:@"46f582f2-b45c-44f2-9723-dbbdc9b07f70"];

	// The first writable (read/write) characteristic exposed by the DTAG100 that we will use to demonstrate 
	// writing via Bluetooth
    self.ourWrite1CharacteristicUUID = [CBUUID UUIDWithString:@"46f582f2-b45c-44f2-9723-dbbdc9b07f81"];

	// Example user friendly names for our regions (these can be used to display notifications when the 
	// user has entered or left a region).
    self.beaconIdentifiers = [NSArray arrayWithObjects:
       @"Beacon 1",
       @"Beacon 2",
       @"Beacon 3",
       @"Beacon 4"
       @"Beacon 5",
       @"Beacon 6",
       @"Beacon 7",
       @"Beacon 8",
       @"Beacon 9",
       nil];
    
	// Example beacon UUIDs to monitor and range (these should match those that the DTAG is configured to broadcast).
    self.beaconUUIDs = [NSArray arrayWithObjects:
                            @"00010203-0405-0607-0809-0A0B0C0D0E01",
                            @"00010203-0405-0607-0809-0A0B0C0D0E02",
                            @"00010203-0405-0607-0809-0A0B0C0D0E03",
                            @"00010203-0405-0607-0809-0A0B0C0D0E04",
                            @"00010203-0405-0607-0809-0A0B0C0D0E05",
                            @"00010203-0405-0607-0809-0A0B0C0D0E06",
                            @"00010203-0405-0607-0809-0A0B0C0D0E07",
                            @"00010203-0405-0607-0809-0A0B0C0D0E08",
                            @"00010203-0405-0607-0809-0A0B0C0D0E09",
                            nil];

	// The UUIDs of the read-only GATT characteristics exposed by the DTAG100
	// The DTAG100 will split the configured readAtt value across up to 8 32 byte
	// characteristics. This example reads these characteristics into an array 
	// (as they may not necessarily arrive in order) and, when all have been read,
	// concatenates them together to form a single value of up to 256 characters.
    self.readCharacteristicUUIDs = [NSArray arrayWithObjects:
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f71",
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f72",
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f73",
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f74",
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f75",
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f76",
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f77",
                              @"46f582f2-b45c-44f2-9723-dbbdc9b07f78",
                              nil];
    
    self.readCharacteristicValues = [NSMutableArray array];
    
	// Initialise the array that will receive the read characteristic values
    for (int i = 0; i < self.readCharacteristicUUIDs.count; i++) {
        [self.readCharacteristicValues addObject:@""];
    }
  
	// Define notification messages in order that selectors can be called indirectly from Bluetooth events
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(launchURL:) name:@"LaunchURL" object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(triggerConnection:) name:@"TriggerConnection" object:nil];

	// Register our beacon regions with the CoreLocation monitoring and ranging
    [self loadRegions];
}
    
// This should be called after any change to the region UUIDs to be monitored/ranged
- (void)loadRegions
{  
	// Stop monitoring and ranging existing regions
	[self stopMonitoringBeacons];
	[self stopRangingBeacons];

	self.regions = [[NSMutableArray alloc] init];

	// In this example we will define 9 beacons (and their associated region names) to be monitored and ranged
    for (int i = 0; i < 9; i++) {
        DOTBeacon *beacon = [self.digitalSignageBeacons objectAtIndex:i];
        NSUUID *uuid = [[NSUUID alloc] initWithUUIDString:beacon.uuid];

        CLBeaconRegion *region = [[CLBeaconRegion alloc] initWithProximityUUID:uuid identifier:beacon.name];
        
        if (region)
        {
            region.notifyOnEntry = YES;
            region.notifyOnExit = YES;
            region.notifyEntryStateOnDisplay = YES;
            
            [self.locationManager startMonitoringForRegion:(CLRegion *) region];
            [self.locationManager startRangingBeaconsInRegion:region];
            
			// Add this region to the array of regions that we are currently monitoring
			// This can be used to stop ranging 
            [self.regions addObject:region];
        }
    }
}


- (void)startMonitoringRegions
{
    int count = (int) [self.regions count];
    for (int i = 0; i < count; i++) {
		if (self.regions[i])
        {
			[self.locationManager startMonitoringForRegion:self.regions[i]];
        }
    }
}

- (void)stopMonitoringRegions
{
	if (self.regions) {
		int count = (int) [self.regions count];
		for (int i = 0; i < count; i++) {
			if (self.regions[i])
			{
				[self.locationManager stopMonitoringForRegion:self.regions[i]];
			}
		}
	}
}

- (void)startRangingBeacons
{
	if (self.regions) {
		int count = (int) [self.regions count];
		for (int i = 0; i < count; i++) {
		   if (self.regions[i]) {
				[self.locationManager startRangingBeaconsInRegion:self.regions[i]];
		   }
		}
	}
}

- (void)stopRangingBeacons
{
	if (self.regions) {
		int count = (int) [self.regions count];
		for (uint i = 0; i < count; i++) {
			if (self.regions[i]) {
				[self.locationManager stopRangingBeaconsInRegion:self.regions[i]];
			}
		}
	}
}

#pragma mark - Location manager delegate

- (void)locationManager:(CLLocationManager *)manager didDetermineState:(CLRegionState)state forRegion:(CLRegion *)region
{
	// CoreLocation will call this when a monitored region has been entered or exited.

    // A user can transition in or out of a region while the application is not running. 
	// When this happens, CoreLocation will launch the application momentarily, call this 
	// delegate method and we will let the user know via a local notification.
    UILocalNotification *notification = [[UILocalNotification alloc] init];
    
    if (state == CLRegionStateInside)
    {
        NSString *message = [NSString stringWithFormat:@"%@ found", region.identifier];
        notification.alertBody = NSLocalizedString(message, @"");
        [self startRangingBeacons];
        NSLog(@"Entered region: %@", region.identifier);
    }
    else if (state == CLRegionStateOutside)
    {
	    NSString *message = [NSString stringWithFormat:@"%@ e", region.identifier];
        notification.alertBody = NSLocalizedString(message, @"");
        [self stopRangingBeacons];
        NSLog(@"Exited region: %@", region.identifier);
    }
    else
    {
        return;
    }
    
    // If the application is in the foreground, it will get a callback to application:didReceiveLocalNotification:.
    // If it's not, iOS will display the notification to the user.
}

- (void)locationManager:(CLLocationManager *)manager didRangeBeacons:(NSArray *)beacons inRegion:(CLBeaconRegion *)region
{
    // CoreLocation will call this delegate method at 1 Hz with updated range information on regions being ranged.
    for (CLBeacon *beacon in beacons) { 
		// This is where the app can display some information about the range of the beacon or display an image to match the 
		// beacon found etc.
		// For this example, we'll just log the ranging information.    
        NSLog(@"Beacon UUID %@, Major: %@, Minor: %@, Range: %.2fm", [beacon.proximityUUID UUIDString], beacon.major, beacon.minor, beacon.accuracy);
        
		// And, for example, set the view controller's view's background color to roughly indicate distance.
        UIColor *col = [[UIColor alloc]initWithRed:1.0/beacon.accuracy green:beacon.accuracy/10 blue:beacon.accuracy/30 alpha:1];
        [self.view setBackgroundColor:col];
        
        NSUUID *firstBeaconUUID = [[NSUUID alloc] initWithUUIDString:[self.beaconUUIDs objectAtIndex:0]];
        
		// For this example, if this beacon corresponds to the first UUID in our array and if the major and minor, let's try to trigger a
		// connection to the DTAG100 and read and write to its characteristics once we have been within approx 1 metre of the beacon
		// for more than 10 seconds.
        if ([beacon.proximityUUID isEqual:firstBeaconUUID] && beacon.major.shortValue == 1 && beacon.minor.shortValue == 1) {
            
			// Have we been approx within 1 metre of the beacon for more than 10 seconds?
			// Note that the beacon.accuracy may be -1 if the beacon disappears unexpectedly      
            if (beacon.accuracy >= 0 && beacon.accuracy <= 1.0) {      
                if (!self.triggered) {
                    if (self.waitingToTrigger) {
                        if ([self.timeOfEntry timeIntervalSinceNow] < -10) {

							// Create a notification to call the triggerConnection selector indirectly
                            [[NSNotificationCenter defaultCenter] postNotificationName:@"TriggerConnection" object:nil];
                            self.waitingToTrigger = false;
                            self.triggered = true;
                        }
                    }
                    else {
                        // Start waiting to trigger
                        self.timeOfEntry = [[NSDate alloc] init];
                        self.waitingToTrigger = true;
                    }
                }
            }
            else {  
                // Outside range for more than 10 seconds?
                if (self.triggered && !self.waitingToTrigger) {

                    // Set the waitingToTrigger flag in order to reset the trigger after a period
                    // of being outside the range.
                    self.waitingToTrigger = true;
                    self.timeOfEntry = [[NSDate alloc] init];
                }
                else if (self.waitingToTrigger && [self.timeOfEntry timeIntervalSinceNow] < -10) {
                    // Reset to allow re-triggering
                    self.triggered = false;
                    self.waitingToTrigger = false;
                }
            }
        }
    }
}

#pragma mark - Notifications 

- (void)triggerConnection:(NSNotification *)notification
{
    if (self.scanning) {
        [self.centralManager stopScan];
    }
    NSlog(@"Scanning for peripherals")];
    self.scanning = true;

	// When we successfuly discover a peripheral and find its services and characteristics
	// we will attempt to write to one of those characteristics
    self.writeData = true;
    
	// Up to this point we have been using CoreLocation to monitor and range beacons.
	// Now we need to use CoreBluetooth to discover a peripheral, connect to it and potentially
	// read and write to its GATT service characteristics.
	// Unfortunately, iOS doesn't give us any way to tie the region information that we have been
	// monitoring and ranging with the peripheral that we want to connect to. 
	// For this reason, the DTAG100 will populate the BLE scan response data with an advertisement
	// for the custom service (by UUID) that it implements. This allows us to request that CoreBluetooth
	// only scans for peripherals that advertise this specific service and so avoid enumerating all 
	// Bluetooth devices that are in range.
	// Additionally, the DTAG100 scan response may also include a dynamic local name value which can
	// further be used to ensure that we only connect to DTAG100 devices that have been configured to
	// broadcast that a particular local name.
    [self.centralManager scanForPeripheralsWithServices:@[self.ourServiceUUID] options:nil];
}


- (void) launchURL:(NSNotification *)notification {

	// This selector is called from a notification when characteristics have been read from the DTAG100
	// GATT service and concatenated to form a URL.
	// This example simply launches that URL for demonstation purposes.
    NSString *urlString = notification.object;

    NSURL *url = [NSURL URLWithString:urlString];
    UIApplication *app = [UIApplication sharedApplication];
    if ([app canOpenURL:url]) {
        [app openURL:url];
        NSLog(@"Opened URL %@", url);
    }
    else {
        NSLog(@"Can't open URL %@", url);
    }
}

#pragma mark - CoreBluetooth Central Manager delegate

- (void)centralManagerDidUpdateState:(CBCentralManager *)central {
    self.cBReady = false;
    switch (central.state) {
        case CBCentralManagerStatePoweredOff:
            NSLog(@"CoreBluetooth BLE hardware is powered off");
            break;
        case CBCentralManagerStatePoweredOn:
            NSLog(@"CoreBluetooth BLE hardware is powered on and ready");
            self.cBReady = true;
            break;
        case CBCentralManagerStateResetting:
            NSLog(@"CoreBluetooth BLE hardware is resetting");
            break;
        case CBCentralManagerStateUnauthorized:
            NSLog(@"CoreBluetooth BLE state is unauthorized");
            break;
        case CBCentralManagerStateUnknown:
            NSLog(@"CoreBluetooth BLE state is unknown");
            break;
        case CBCentralManagerStateUnsupported:
            NSLog(@"CoreBluetooth BLE hardware is unsupported on this platform");
            break;
        default:
            break;
    }
}


- (void)centralManager:(CBCentralManager *)central
  didDiscoverPeripheral:(CBPeripheral *)peripheral
      advertisementData:(NSDictionary *)advertisementData
                   RSSI:(NSNumber *)RSSI {

	// This is called by CoreBluetooth when scanning for peripherals 
	// In our example, this is restricted to peripherals which advertise the DTAG100
	// custom GATT service.
    NSLog(@"Discovered %@", peripheral.name);

	// We can get additional information from the advertisement and scan response data from
	// the dictionary. 
    NSArray *services = [advertisementData objectForKey:CBAdvertisementDataServiceUUIDsKey];
    NSLog(@"Services advertised: %@", services);
    
    NSString *localName = [advertisementData objectForKey:CBAdvertisementDataLocalNameKey];
    NSLog(@"Local name: %@", localName);

	//Unfortunately, the manufacturer data seems to be empty (at least on current
	// versions of iOS) otherwise we would be able to compare it with the beacon ID which 
	// we were ranging.
	// NSData *manDataNS = [advertisementData objectForKey:CBAdvertisementDataManufacturerDataKey];
  
	// We can compare the local name here if we need to only connect to particular DTAG100 devices.
	NSString *localNameToMatch = @"DTAG100";
    if ([localName isEqual:localNameToMatch]) {
        // We've found our device, stop scanning, keep a copy of the peripheral and connect to it
        [self.centralManager stopScan];
        self.scanning = false;
        [self.centralManager connectPeripheral:peripheral options:nil];
        self.peripheral = peripheral;
    }
}

- (void)centralManager:(CBCentralManager *)central didConnectPeripheral:(CBPeripheral *)peripheral {
    // This is called by CoreBluetooth when a connection has been established.
    NSLog(@"Peripheral connected");
    peripheral.delegate = self;
     
    // Discover the services and read/write the characteristics and disconnect as quickly as possible
	// As we know our service UUID, we can specifically look for that service only
    [self.peripheral discoverServices:@[self.ourServiceUUID]];
}

- (void)centralManager:(CBCentralManager *)central didFailToConnectPeripheral:(CBPeripheral *)peripheral error:(NSError *)error {
    NSLog(@"Failed to connect");
    [self cleanup];
}

- (void)cleanup {
	// We call this from any CBCentralManager delegate that indicates an error in order to 
	// cancel the connection. 
	// Any other required cleanup could be done here also.  
    [self.centralManager cancelPeripheralConnection:self.peripheral];
}


- (void)centralManager:(CBCentralManager *)central didDisconnectPeripheral:(CBPeripheral *)aPeripheral error:(NSError *)error
{
	// CoreBluetooth calls this when a peripheral has been disconnected.
	// We can use this as an opportunity to resume monitoring and ranging activities if we have stopped.
    NSLog(@"Disconnected from %@", aPeripheral.name");
  
    if (self.peripheral) {
        [self.peripheral setDelegate:nil];
        self.peripheral = nil;
    }
    
    [self startMonitoringRegions];
    [self startRangingBeacons];
}

#pragma mark - CoreBluetooth Peripheral delegate

- (void)peripheral:(CBPeripheral *)peripheral didDiscoverServices:(NSError *)error {

	// CoreBluetooth calls this when services have been discovered - this gives
	// us handles to the discovered services that we can use to discover characteristics
    bool found = false;
    
    if (error) {
         NSLog(@"Error discovering services %@", error);
        [self cleanup];
        return;
    }
    
    for (CBService *service in peripheral.services) {
        NSLog(@"Discovered service %@", service.UUID);
        NSLog(@"Comparing to %@", self.ourServiceUUID);
        
		// Verify that we have discovered the custom service UUID that the DTAG100 implements
        if ([service.UUID isEqual:self.ourServiceUUID]) {

			// Enumerate the characteristics that this service implements
            [peripheral discoverCharacteristics:nil forService:service];
            found = true;
        }
    }

    if (!found) {
        // Disconnect
        [self cleanup];
    }
}

- (void)peripheral:(CBPeripheral *)peripheral didDiscoverCharacteristicsForService:(CBService *)service error:(NSError *)error {
	
	// CoreBluetooth calls this when characteristics have been discovered for a service
	// This gives us handles to the characteristics that we can use to read and write
    if (error) {
        NSLog(@"Error discovering characteristics: %@", [error localizedDescription]);
        [self cleanup];
        return;
    }
    
    self.characteristicsCount = 0;
    self.readCount = 0;
    uint count = 0;
    
	// The DTAG100 implements a custom GATT service comprises 8 read-only characteristics
	// and 4 or more read/write characteristics
    for (CBCharacteristic *characteristic in service.characteristics) {
        NSLog(@"Discovered characteristic %@", characteristic.UUID);
        
        NSLog(@"Reading value for characteristic %@", characteristic.UUID);

		// Initiate a read of this characteristic's value
        [peripheral readValueForCharacteristic:characteristic];
        count++;
        
        NSMutableString *string;
        
		// If this is the DTAG100's first read/write characteristic, we can write a message to it.
        if (self.writeData && [characteristic.UUID isEqual:self.ourWrite1CharacteristicUUID]) {
          
			// Write a simple example message to the read/write characteristic
            NSString *stringToWrite = @"Hello, DTAG100";
			NSLog(@"Writing value %@ to characteristic %@", self.takeAwaySettings.writeTextInit, characteristic.UUID);
            NSData* data = [stringToWrite dataUsingEncoding:NSASCIIStringEncoding];
            [peripheral writeValue:data forCharacteristic:characteristic type:CBCharacteristicWriteWithResponse];
        }
    }
    
	// Save this count to indicate how many characteristics were requested to be read 
    self.characteristicsCount = count;
    
    if (count == 0) {
        NSLog(@"No characteristics to read");
        [self cleanup];
    }
    
}

- (void)peripheral:(CBPeripheral *)peripheral didWriteValueForCharacteristic:(CBCharacteristic *)characteristic
            error:(NSError *)error {
              
	// CoreBluetooth calls this when characteristics value has been WRITTEN (instigated by a call to writeValue) 		                               
    if (error) {
        NSLog(@"Error writing characteristic value: %@", [error localizedDescription]);
        [self cleanup];
        return;
    }
    
    NSLog(@"Write complete");
}

- (void)peripheral:(CBPeripheral *)peripheral didUpdateValueForCharacteristic:(CBCharacteristic *)characteristic
             error:(NSError *)error {
    
	// CoreBluetooth calls this when characteristics value has been READ (instigated by a call to 
	// readValueForCharacteristic).
	// Note that the characteristics are not necessarily returned in the order that they were requested so it is
	// necessary to check for which characteristic this is the value.
            
    if (error) {
        NSLog(@"Error discovering characteristics: %@", [error localizedDescription]);
        [self cleanup];
        return;
    }
    
    NSData *data = characteristic.value;
    NSString *valueString = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
    NSLog(@"Read value for characteristic %@:%@", characteristic.UUID, valueString);
    self.readCount++;
    
    // Populate the appropriate read characteristic value in our array based on the characteristic's UUID
    bool found = false;
    for (int i = 0; !found && i < self.readCharacteristicUUIDs.count; i++) {
        CBUUID *uuid = [CBUUID UUIDWithString:[self.readCharacteristicUUIDs objectAtIndex:i]];
        
        if ([characteristic.UUID isEqual:uuid]) {
            self.readCharacteristicValues[i] = valueString;
            found = true;
        }
    }
    
    if (self.characteristicsCount > 0 && self.characteristicsCount <= self.readCount) {
		// All of the characteristic values that we have requested have now been read so we can 
		// disconnect as fast as possible to make way for another connection.
        NSLog(@"All characteristics read");
        self.readCount = 0;
        [self cleanup];
        
        // Construct the read value by concatenating all of the read characteristic values in order
		// This creates a single string of up to 256 characters which, in this example we will use 
		// to launch a URL.
        self.launchValue = [[NSMutableString alloc] initWithString:@""];
        for (NSString *value in self.readCharacteristicValues) {
            [self.launchValue appendString:value];
        }
             
		// Post a notification to call the launchURL selector with our launchValue string
        [[NSNotificationCenter defaultCenter] postNotificationName:@"LaunchURL" object:self.launchValue];
    }
}

@end
